/* 
   tilt_p3.c  -  (c) 1999 Till Harbaum

   tilt sensor driver for old style palm tilt sensor

   t.harbaum@tu-bs.de

   This program is GPL'd freeware.
*/

#include <SysAll.h>
#include <Pilot.h>
#include <System/KeyMgr.h>

#define BYTEREG(a)  (*(volatile unsigned char*)a)
#define PEDIR       BYTEREG(0xfffff420)
#define PEDATA      BYTEREG(0xfffff421)
#define PEUP        BYTEREG(0xfffff422)
#define PESEL       BYTEREG(0xfffff423)
#define PJDIR       BYTEREG(0xfffff438)
#define PJDATA      BYTEREG(0xfffff439)
#define PJSEL       BYTEREG(0xfffff43b)

#define LONGREG(a)  (*(volatile unsigned long*)a)
#define PORTE       LONGREG(0xfffff420)
#define PORTJ       LONGREG(0xfffff438)

/* interrupt mask register */
#define IMR  (*(unsigned long*)0xfffff304)

static int tilt_check_cpu(void) {
  /* if we're not using a new include file, define things appropriately */
#ifndef sysFtrNumProcessorID
#define sysFtrNumProcessorID    2
  /* Product id */
  /* 0xMMMMRRRR, where MMMM is the processor model */
  /* and RRRR is the revision. */
#define sysFtrNumProcessorMask  0xFFFF0000 // Mask to obtain processor model
#define sysFtrNumProcessor328   0x00010000 // Motorola 68328 (Dragonball)
#define sysFtrNumProcessorEZ    0x00020000 // Motorola 68EZ328 (Dragonball EZ)
#endif
  DWord id, chip;
  Err err;

  err = FtrGet(sysFtrCreator, sysFtrNumProcessorID, &id);
  if (err) {
    // can't get that feature; we must be on an old unit
    // without it defined, thus we're on a DragonBall.
    chip = sysFtrNumProcessor328;
  } else
    chip = id & sysFtrNumProcessorMask;

  return(chip == sysFtrNumProcessor328);
}

void getvals(short *vals);

asm("
getvals:
        link.w    %a6,#0
	movem.l   %d0-%d5/%a1,-(%a7)
	
        move.l    8(%a6),%a0
	move.l	  #0xfffff421, %a1  | Port E data reg

	move	  #2000,%d0  | init div counter
	move	  %d0,%d1
	move	  %d1,%d2
	move	  %d2,%d3
	move	  %d3,%d4
	move	  %d4,%d5
		
loop_0:	btst.b    #6,(%a1)	| wait while x high
	dbeq	  %d0,loop_0

loop_1:	btst.b    #6,(%a1)	| wait while x low
	dbne	  %d1,loop_1

	| this is time Ta
	
loop_2:	btst.b    #6,(%a1)	| wait while x high
	dbeq	  %d2,loop_2

	| this is time Tb

loop_3:	btst.b    #5,(%a1)	| wait while y high
	dbeq	  %d3,loop_3
		
loop_4:	btst.b    #5,(%a1)	| wait while y low
	dbne	  %d4,loop_4
	| this is time Tc
	
loop_5:	btst.b    #5,(%a1)	| wait while y high
	dbeq	  %d5,loop_5

	| this is time Td

	move.w    %d0,(%a0)+
	move.w    %d1,(%a0)+
	move.w    %d2,(%a0)+
	move.w    %d3,(%a0)+
	move.w    %d4,(%a0)+
	move.w    %d5,(%a0)+

	movem.l   (%a7)+,%d0-%d5/%a1
	unlk      %a6
	rts
");


/* switch power pin on/off */
void tilt_power(Boolean on) { 
  if(on) PJDATA |=  0x80;
  else   PJDATA &= ~0x80;
}

/* switch self test pin on/off */
void tilt_selftest(Boolean on) {
  if(on) PJDATA |=  0x40;
  else   PJDATA &= ~0x40;
}

void tilt_setup(void) {

  /* Port E */
  /* set upper two bits to gpio */
  PESEL |= 0x60; 

  /* disable pullups */
  PEUP  &= ~0x60; 

  /* set both pins to input */
  PEDIR &= ~0x60;

  /* Port J */
  /* set upper two bits to gpio */
  PJSEL |= 0xc0; 

  /* set both pins to output */
  PJDIR |=  0xc0;

  /* both output pins low (chip off) */
  PJDATA &= ~0xc0;

  /* now switch on chip */
  tilt_power(1);
  tilt_selftest(0);
}

#define SAMPLE 8
Boolean tilt_measure(Word *x, Word *y) {
  unsigned long irq_save;
  unsigned short i,j;
  int ax=0,ay=0;
  int t2[SAMPLE];
  short a[SAMPLE][6];

  /* disable all irqs */
  irq_save = IMR;
  IMR = 0x00ffffff;

  for(i=0;i<SAMPLE;i++)
    getvals(a[i]);

  /* reenable irqs */
  IMR = irq_save;

  /* counter went down to 0?? -> no chip found */
  if((a[0][0]==-1)||(a[0][1]==-1)) return false;

  for(i=0;i<SAMPLE;i++) {
    for(j=0;j<6;j++)  
      a[i][j] = 2000-a[i][j];

    /* calculate T2 */
    t2[i] = a[i][2]/2 + a[i][3] + a[i][4] + a[i][5]/2;

    ax += ((unsigned long)a[i][2]<<8)/t2[i];
    ay += ((unsigned long)a[i][5]<<8)/t2[i];
  }

  *x=(ax/SAMPLE); 
  *y=(ay/SAMPLE);

  return true;
}

Err tilt_init(struct globals *globs) {

  /* Initialize the globals */
  globs->zero_x   = 0;
  globs->zero_y   = 0;

  /* is this the right hardware */
  if(!tilt_check_cpu()) return 2;

  /* initialize sensor */
  tilt_setup();

  /* try to access sensor */
  if(!tilt_measure(&globs->zero_x, &globs->zero_y)) 
    return 1;

  /* sensor is fine */
  return 0;
}

/* zero the tilt sensor */
Err tilt_zero(struct globals *globs) {

  /* get zero values */
  tilt_measure(&globs->zero_x, &globs->zero_y);

  return 0;
}

/* return normalized tilt value */
Err tilt_get(struct globals *globs, Word *x, Word *y) {
  tilt_measure(x,y);
  *x -= globs->zero_x;
  *y -= globs->zero_y;

#ifdef DEBUG
  {
    char str[32];
    StrIToA(str, *x);
    WinDrawChars(str, StrLen(str), 0,0);
  }
#endif

  return 0;
}

/* get tilt prefs */
void tilt_prefs(struct globals *globs) {
  Word size;

  /* get preferences, init prefs if none found */
  size = sizeof(struct globals);
  if(PrefGetAppPreferences(CREATOR, 0, globs, &size, true) ==
     noPreferenceFound) {
    
    /* zero sensor */
    tilt_zero(globs);      
    
    /* default buton mappings */
    globs->map[0] = 0;
    globs->map[1] = 0;
    globs->map[2] = 0;
    globs->map[3] = 0;
    
    /* mapping disabled */
    globs->flags  = 0;

    globs->sensivity = 0;

    PrefSetAppPreferences(CREATOR, 0, 0, globs, sizeof(struct globals), true);
  }
}








